{-# LANGUAGE TypeOperators,
             GADTs,  
             RankNTypes,
             FlexibleContexts,
             FlexibleInstances,
             MultiParamTypeClasses,
             UndecidableInstances #-}

module AOP where

import Control.Monad.Trans
import Control.Monad.Identity
import Control.Monad.State
import Control.Monad.Writer

type Tag = Integer

--------------------------------------------------------------------------------
-- Join point model
--------------------------------------------------------------------------------

-- | A jointpoint is the combination of a function to be identified and a value
data Jp m a b = Monad m => Jp Tag

-- | A pointcut is a predicate over join points
type Pc m a b = Jp m a b -> Bool

-- | An advice transforms a function into another one of the same type
type Advice m a b = (a -> m b) -> (a -> m b)

-- | An aspect pairs a pointcut and an advice
type Aspect m a b = (Pc m a b, Advice m a b)

-- | The aspect enviroment is a list of deployed aspects
-- | We consider only functions with type (Int -> m Int)
type AEnv m = [Aspect m Int Int]

-- | Predefined pcTag pointcut; matches join points with tag t.
pcTag t (Jp t') = t == t'


--------------------------------------------------------------------------------
-- Weaver
--------------------------------------------------------------------------------

weave :: Monad m =>  (Int -> m Int) -> AEnv m ->
                     Jp m Int Int -> m (Int -> m Int)
weave f [] _                 = return f
weave f ((pc, adv):asps) jp  = 
  weave (if pc jp then (adv f) else f) asps jp


--------------------------------------------------------------------------------
-- Aspectual Monad Transformer
--------------------------------------------------------------------------------

data AOT m a = AOT { unAOT :: m (ResultAOT m a) } 

data  ResultAOT m a
      =  Return a
      |  OpenApp  Tag                    -- tag
                  (Int -> AOT m Int)     -- function
                  Int                    -- argument
                  (Int -> AOT m a)       -- continuation

instance Monad m => Monad (AOT m) where
  return   =   AOT . return . Return
  m >>= f  = AOT (unAOT m >>= \r -> case r of
     Return x        -> unAOT (f x)
     OpenApp t g x k   ->
        return (OpenApp t g x (\y -> k y >>= f)))

instance MonadTrans AOT where
  lift ma = AOT (ma >>= \ a -> (return . Return) a)    

openApp t f x = AOT (return (OpenApp t f x return))

runAOT aenv m = unAOT m >>= go where
  go (Return r)       = return r
  go (OpenApp t f x k)  =
         unAOT (  weave f aenv (Jp t) >>= 
            \woven_f -> woven_f x >>= k) >>= go

-- | Type class for open applications

class Monad m => AOPMonad m where
  wapp :: Tag -> (Int -> m Int) -> (Int -> m Int)

instance Monad m => AOPMonad (AOT m) where 
  wapp t f x = openApp t f x

-- | Compatibility with standard effects: state and writer

instance MonadState s m => MonadState s (AOT m) where
  put = lift . put 
  get = lift $ get  

instance MonadWriter s m => MonadWriter s (AOT m) where
  tell = lift .tell 
